/*
 * Bang Head Here - Control
 *
 * Copyright (C) 2016  B.Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <avr/wdt.h>

#include "Adafruit_NeoPixel.h"
#include "notes.h"

//#define DEBUG	1

#define NELEM(x)	(sizeof(x) / sizeof((x)[0]))

/*
 * Port mapings:
 *	Port B:
 *		0 - unused
 *		1 - unused
 *		2 - (OC1B) Tone
 *		3 - unused
 *		4 - Buzzer pull-up
 *		5 - Hands
 *	Port C:
 *		0 - (D) Clear peak high
 *		1 - (D) Clear peak low
 *		2 - (A) Peak level high
 *		3 - (A) Peak level low
 *		4 - unused
 *		5 - unused
 *		6 - unused
 *		7 - unused
 *	Port D:
 *		0 - (RXD)
 *		1 - (TXD)
 *		2 - WS2812 strip data
 *		3 - (OC2B) Red PWM
 *		4 - unused
 *		5 - (OC0B) Green PWM
 *		6 - (OC0A) Blue PWM
 *		7 - unused
 */
/* The analog bits are on PORTC */
#define BIT_PEAKL	3	// Low level peak analog input
#define BIT_PEAKH	2	// High level peak analog input
#define BIT_CLRL	1	// Clear low level peak detector (HiZ, 1)
#define BIT_CLRH	0	// Clear high level peak detector (HiZ, 0)
#define BIT_TONE	2	// PORTB bit 2 (OC1B, Arduino digital pin 10)

#define PIN_PEAKL	A3	// PORTC bit 3
#define PIN_PEAKH	A2	// PORTC bit 2
#define PIN_CLRL	A1	// PORTC bit 1
#define PIN_CLRH	A0	// PORTC bit 0
#define PIN_HANDS	13	// PORTB bit 5 - Hand-plates input (LED removed from arduino)
#define PIN_PULL	12	// PORTB bit 4 - Buzzer pull-up
#define PIN_TONE	10	// PORTB bit 2 - Buzzer tone generator output
#define PIN_LEDB	6	// PORTD bit 6 - Plate RGB led Blue
#define PIN_LEDG	5	// PORTD bit 5 - Plate RGB led Green
#define PIN_LEDR	3	// PORTD bit 3 - Plate RGB led Red
#define PIN_WSDATA	2	// PORTD bit 2 - WS2812B led strip force indicator

#define PIN_UNUSED4	4	// PORTD bit 4
#define PIN_UNUSED7	7	// PORTD bit 7
#define PIN_UNUSED8	8	// PORTB bit 0
#define PIN_UNUSED9	9	// PORTB bit 1
#define PIN_UNUSED11	11	// PORTB bit 3

#define NPSTRIP		19	// Pixels in line (first in chain)
#define NPCROSS		4	// Pixels in cross-bar (last in chain)
#define NPIXELS		(NPSTRIP + NPCROSS)

#define waitforzero(reg,bit)	do { \
					while((reg) & _BV(bit)) \
						; \
				} while(0)
#define waitforone(reg,bit)	do { \
					while(!((reg) & _BV(bit))) \
						; \
				} while(0)

/* Tone pitch table for 16MHz crystal */
static const uint16_t tone_pitch[NTONE_PITCH] PROGMEM = {
	61155, 57722, 54482, 51424, 48538, 45814, 43242, 40815, 38524, 36362 /* == 110Hz */, 34321, 32395,
	30577, 28860, 27240, 25711, 24268, 22906, 21620, 20407, 19261, 18180 /* == 220Hz */, 17160, 16197,
	15288, 14429, 13619, 12855, 12133, 11452, 10809, 10203,  9630,  9089 /* == 440Hz */,  8579,  8098,
	 7643,  7214,  6809,  6427,  6066,  5725,  5404,  5101,  4814,  4544,  4289,  4048,
	 3821,  3606,  3404,  3213,  3032,  2862,  2701,  2550,  2406,  2271,  2144,  2023,
	 1910,
};

Adafruit_NeoPixel bangpower = Adafruit_NeoPixel(NPIXELS, PIN_WSDATA, NEO_GRB | NEO_KHZ800);

#define TIME_HANDS		25	// 25 ms debounce time after last change detected
#define TIME_PEAK		4	// 4 ms delay for centering the voltage
#define TIME_IDLEANIM		50	// Idle animation interval
#define TIME_PEAKMEASURE	5000	// Time to wait for a peak
#define TIME_PEAKEXTEND		100	// Time after a peak has been seen to get full extend
#define TIME_PLATEBLINK		166	// Blink plate at 3Hz
#define TIME_PLATEBLINKFAIL	83	// Blink plate at 6Hz
#define TIME_PEAKANIM		66	// Peak animation interval
#define TIME_WATCHDOG		75	// 75ms kick dog interval for a 250ms dog reset
#define TIME_WINANIM		25	// Win animation interval

#define EV_TIMER	0x01	// Generic timer expired
#define EV_TMRTONE	0x02	// Sound tone-timer expired
#define EV_TMRPEAK	0x04	// Peak enable delay expired
#define EV_TMRBLINK	0x08	// Blink timer expired
#define EV_HANDS_OFF	0x10	// Hands left the plates
#define EV_HANDS_ON	0x20	// Hands touch the plates

#define PEAKMIN		90	// Least level of banging to do something
#define PEAKMAX		860	// Hardest bang to indicate top-level

#define NFAILBLINK	20	// Number of times to blink on failed to bang on plate in time

enum {
	ST_IDLE,		// Default state, idle animations
	ST_HANDS_ON,		// Hands are placed, beep
	ST_WAIT_TONE,		// Wait for beep to finish
	ST_WAIT_PEAKENABLE,	// Wait for stable peak mesurement possible
	ST_PEAKMEASURE,		// Measure the Bang!
	ST_PEAKFAIL,		// No bang within timeout
	ST_PEAKWAIT,		// Seen a bang, let it fade
	ST_PEAKDONE,		// Animate the magnitude of the bang
	ST_WAIT_HANDS_OFF,	// Wait for the hands to be removed
	ST_WINNER,		// Perfect Bang!
};

static uint8_t prevtime;	// Previous timestamp of loop
static uint8_t state;		// State of game
static uint8_t hands;		// State of hands on input (low == hands on the plates)
static uint8_t handsold;	// Previous hands state for debouncing
static uint16_t timer_tone;	// Tone generator timing
static uint8_t timer_hands;	// Debounce timer
static uint8_t timer_peak;	// Peak-detector enable delay timer
static uint16_t timer_blink;	// Blink timer
static uint16_t timer;		// Generic timer
static uint8_t timer_dog;	// Watchdog timer
static int16_t peakl;		// Actual peak values
static int16_t peakh;
static int16_t peakpp;		// Maximum p-p value of input
static uint8_t blinkstate;	// Keep track whether the bangplate is lit or not

/*
 * Stop the buzzer from buzzing
 */
static inline void sound_off(void)
{
	TCCR1B &= ~_BV(CS10);		// Disable timer
	TCCR1A = 0;			// Disconnect compare match
	// The output is already set high and will automatically enable the
	// buzzer low-side driver-transistor for bang-detection.
	digitalWrite(PIN_PULL, 1);	// Disable the high-side driver
}

/*
 * Set the speaker to generate a tone
 */
static inline void sound_tone(uint8_t note, uint16_t duration)
{
	timer_tone = duration;
	if(note && note <= NTONE_PITCH) {
		digitalWrite(PIN_PULL, 0);	// Enable the high-side driver
		ICR1 = pgm_read_word(&tone_pitch[note-1]);
		TCCR1A = _BV(COM1B0);		// Set toggle output
		TCNT1 = 0;
		TCCR1B |= _BV(CS10);		// Enable timer
	} else {
		sound_off();
	}
}

/*
 * Sound timer initialization
 *
 * Setup sound hardware with Timer1 manually. The disabled interrupts
 * in the pixel-update causes interference in the timing resulting in
 * clicks in the sound output.
 */
static inline void sound_init(void)
{
	ICR1 = NOTE_C4;			// Just a value
	OCR1A = 1;			// compare match
	OCR1B = 1;			// compare match
	TCCR1A = 0;			// Disconnect output match
	TCCR1B = _BV(WGM13) | _BV(WGM12);	// CTC mode
	TCNT1 = 0;
}

/*
 * Winning tune
 * (part of Pippi Långstrump theme tune)
 */
static uint8_t tune_idx;
static const struct {
	uint8_t		note;
	uint16_t	duration;
} tune_win[] PROGMEM = {
	{NOTE_C4, DUR_1_8},
	{NOTE_F4, DUR_1_8},
	{NOTE_A4, DUR_1_8},
	{NOTE_F4, DUR_1_8},
	{NOTE_G4, DUR_1_4},
	{NOTE_As4, DUR_1_16},
	{NOTE_A4, DUR_1_16},
	{NOTE_G4, DUR_1_16},
	{NOTE_F4, DUR_1_16},
	{NOTE_E4, DUR_1_8},
	{NOTE_G4, DUR_1_8},
	{NOTE_C4, DUR_1_8},
	{NOTE_E4, DUR_1_8},
	{NOTE_F4, DUR_1_4},
	{NOTE_A4, DUR_1_4},

	{NOTE_C4, DUR_1_8},
	{NOTE_F4, DUR_1_8},
	{NOTE_A4, DUR_1_8},
	{NOTE_F4, DUR_1_8},
	{NOTE_G4, DUR_1_4},
	{NOTE_As4, DUR_1_16},
	{NOTE_A4, DUR_1_16},
	{NOTE_G4, DUR_1_16},
	{NOTE_F4, DUR_1_16},
	{NOTE_E4, DUR_1_8},
	{NOTE_G4, DUR_1_8},
	{NOTE_C4, DUR_1_8},
	{NOTE_E4, DUR_1_8},
	{NOTE_F4, DUR_1_2},
};

/*
 * Next note in the winner tune
 */
static uint8_t sound_tune_next(void)
{
	if(tune_idx < NELEM(tune_win)) {
		uint16_t dur = pgm_read_word(&tune_win[tune_idx].duration);
		uint8_t note = pgm_read_byte(&tune_win[tune_idx].note);
		sound_tone(note, dur);
		tune_idx++;
		return 1;
	} else
		return 0;
}

/*
 * Start playing the winner tune
 */
static void sound_tune_start(void)
{
	tune_idx = 0;
	sound_tune_next();
}

/*
 * Test if ADC is done converting
 */
static inline uint8_t adc_isready(void)
{
	return 0 == (ADCSRA & _BV(ADSC));
}

/*
 *
 */
static inline void adc_measure_start(void)
{
	waitforzero(ADCSRA, ADSC);
	ADMUX ^= _BV(MUX0);		// Change channel (alternate between PEAKL and PEAKH)
	ADCSRA |= _BV(ADSC);		// Start conversion
}

/*
 * Initialize the ADC to our requirements
 */
static inline void adc_init(void)
{
	ADCSRA &= ~_BV(ADEN);		// Disable ADC (enabled in arduino init)
	ADCSRB = 0;			// Single mode conversion
	ADCSRA = _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);	// ADC at 1:128 clock
	ADMUX = _BV(REFS0) | _BV(MUX1);	// Reference AVcc, right adjust, channel 2
	ADCSRA |= _BV(ADEN);		// Enable ADC
	waitforzero(ADCSRA, ADSC);
	ADCSRA |= _BV(ADSC);		// Start first conversion (takes longer)
	waitforzero(ADCSRA, ADSC);	// Wait for it to complete
}

/*
 * Initialize the peak detector I/O pins
 */
static inline void peak_init(void)
{
	PORTC = 0;	// Clear port C, pull-ups disabled
	// Set analog pins to input, unused PC4/PC5 to output(0)
	DDRC = ~(_BV(BIT_CLRL) | _BV(BIT_CLRH) | _BV(BIT_PEAKL) | _BV(BIT_PEAKH));
	// Disable the input drivers for the used analog pins
	DIDR0 |= _BV(BIT_CLRL) | _BV(BIT_CLRH) | _BV(BIT_PEAKL) | _BV(BIT_PEAKH);
}

/*
 * Clear the peak detector
 */
static inline void peak_clear(void)
{
	PORTC |= _BV(BIT_CLRL);				// Set clear to Vcc
	PORTC &= ~_BV(BIT_CLRH);			// Set clear to GND
	DDRC |= _BV(BIT_CLRL) | _BV(BIT_CLRH);		// Set to output
	// Peak detectors are now reset, taking ~120us
}

/*
 * Prepare for peak detector enable
 *
 * peak_enable() can be called about 3ms after preparation. This is due to the
 * time it takes for the buzzer to level out at 1/2 Vcc. The resistance and
 * capacitance combination has an RC-time of about 500us.
 */
static inline void peak_enable(void)
{
	sound_off();					// Ensure no interference and pin-setup
	digitalWrite(PIN_PULL, 1);			// Disable the high-side driver
	peak_clear();
	timer_peak = TIME_PEAK;
}

/*
 * Enable the peak detector
 * Disables the input buffers to prevent leakage currents
 */
static inline void peak_enable_input(void)
{
	// XXX: peak_enable() must be called before calling enable.
	DIDR0 |= _BV(BIT_CLRL) | _BV(BIT_CLRH);		// Make sure to disable digital buffers
	DDRC &= ~(_BV(BIT_CLRL) | _BV(BIT_CLRH));	// Set to input
	// Peak detectors are available after ~12us
	delayMicroseconds(10);
	peakl = peakh = 0;
	peakpp = 0;
	adc_measure_start();
}

/*
 * Test if we are in peak detection mode
 */
static inline uint8_t peak_isenabled(void)
{
	return !timer_peak || !(DDRC & _BV(BIT_CLRH));
}

/*
 * Timer countdown and return an event on timer reaching zero.
 */
static inline uint8_t do_timer(uint8_t tdiff, uint16_t &tmr, uint8_t ev)
{
	if(tmr) {
		if(tmr <= tdiff) {
			tmr = 0;
			// Set event when done counting
			return ev;
		} else
			tmr -= tdiff;	// Timer countdown
	}
	return 0;
}

static inline uint8_t do_timer(uint8_t tdiff, uint8_t &tmr, uint8_t ev)
{
	if(tmr) {
		if(tmr <= tdiff) {
			tmr = 0;
			// Set event when done counting
			return ev;
		} else
			tmr -= tdiff;	// Timer countdown
	}
	return 0;
}

/*
 * Set the bang-plate color
 */
static void plate_color(uint8_t r, uint8_t g, uint8_t b)
{
	analogWrite(PIN_LEDR, r);
	analogWrite(PIN_LEDG, g);
	analogWrite(PIN_LEDB, b);
}

static void plate_color(uint16_t idx)
{
	uint32_t clr = bangpower.getPixelColor(idx);
	plate_color(clr >> 16, clr >> 8, clr);
}

/*
 * Animation when a bang has seen a peak
 */
static uint8_t peakanim_peakidx;
static uint8_t peakanim_step;
static uint8_t peakanim_view;
static uint32_t peakanim_maxclr;

static void peak_animate_reset(void)
{
	/*
	 * Top level is peakpp.
	 * Map the levels to the LED strip
	 */
	peakanim_peakidx = NPSTRIP * (peakpp - PEAKMIN) / (PEAKMAX - PEAKMIN);
	if(peakanim_peakidx < 1)
		peakanim_peakidx = 1;
	if(peakanim_peakidx > NPSTRIP)
		peakanim_peakidx = NPSTRIP;
	peakanim_step = 1;
	peakanim_view = 0;
#ifdef DEBUG
	Serial.print(F("Peak index: "));
	Serial.println(peakanim_peakidx);
#endif
}

/*
 * Quadratic scale for approx linear intensity visual in the color bar after a
 * Bang.
 */
static const uint8_t peak_intensities[] PROGMEM = {
	  1,   2,   4,  7,  12,  19,  28,  38,
	 50,  63,  78, 95, 113, 133, 154, 177,
	201, 227, 255,
};

static uint8_t peak_animate_step(void)
{
	if(peakanim_step <= peakanim_peakidx) {
		// Set the strip incrementally to the peak point on the strip
		// until it reaches the level recorded.
		bangpower.clear();
		uint8_t i;
		for(i = 0; i < peakanim_step; i++) {
			uint8_t v = pgm_read_byte(&peak_intensities[NELEM(peak_intensities) - peakanim_step + i]);
			// Start at blue and work around the color circle to red
			bangpower.setPixelColorHsv(i, 4 * NEO_HSV_HUE_SEXTANT- 4 * NEO_HSV_HUE_SEXTANT * i / (NPSTRIP-1), 255, v);
		}
		peakanim_maxclr = bangpower.getPixelColor(i-1);
		bangpower.setPixelColor(NPSTRIP+0, peakanim_maxclr);
		bangpower.setPixelColor(NPSTRIP+1, peakanim_maxclr);
		bangpower.setPixelColor(NPSTRIP+2, peakanim_maxclr);
		bangpower.setPixelColor(NPSTRIP+3, peakanim_maxclr);
		plate_color(i-1);
		bangpower.show();
		peakanim_step++;
	} else {
		if(peakanim_peakidx == NPSTRIP/2 + 1) {
			// We have a winner!
#ifdef DEBUG
			Serial.println(F("Winner..."));
#endif
			// Copy the pixel colors from bottom to top in mirror fashion
			for(uint8_t i = 0; i < NPSTRIP/2; i++)
				bangpower.setPixelColor(NPSTRIP-i-1, bangpower.getPixelColor(i));
			bangpower.show();
			state_set(ST_WINNER);
			return 1;
		} else {
			// Blink the top-level reached by the bang
			if(!peakanim_view)
				sound_tone(NOTE_B2, DUR_1_4);
			if(peakanim_view & 2) {
				bangpower.setPixelColor(peakanim_peakidx-1, peakanim_maxclr);
				bangpower.setPixelColor(NPSTRIP+0, peakanim_maxclr);
				bangpower.setPixelColor(NPSTRIP+3, peakanim_maxclr);
			} else {
				bangpower.setPixelColor(peakanim_peakidx-1, 0);
				bangpower.setPixelColor(NPSTRIP+0, 0);
				bangpower.setPixelColor(NPSTRIP+3, 0);
			}
			bangpower.show();
			peakanim_view++;
		}
	}
	return peakanim_view < 16;
}

enum {
	ST_IDLE_ANIM_BLANK,
	ST_IDLE_ANIM_RANDOM,
	ST_IDLE_ANIM_PINGPONG,
	ST_IDLE_ANIM_RAINBOW,
	ST_IDLE_ANIM_BAR,
	ST_IDLE_ANIM_GLOW,
	ST_IDLE_ANIM_MAX,
};

/*
 * Bi-quadratic scale for approx linear intensity visual in the B/W idle glow
 * of the place.
 */
static const uint8_t glow_scale[] PROGMEM = {
	0, 1, 2, 3, 4, 7, 9, 13, 16, 21, 25, 31, 36, 43, 49, 57, 64, 73, 81, 91, 100, 111, 121, 133, 144, 157, 169, 183, 196, 211, 225, 241,
	255, 241, 225, 211, 196, 183, 169, 157, 144, 133, 121, 111, 100, 91, 81, 73, 64, 57, 49, 43, 36, 31, 25, 21, 16, 13, 9, 7, 4, 3, 2, 1,
};

#define IDLE_BLANK_CNT		100	// About 5 seconds
#define IDLE_RANDOM_CNT		60
#define IDLE_RAINBOW_CNT	100
#define IDLE_PINGPONG_CNT	((NPSTRIP/2 + 3) * 12)
#define IDLE_BAR_CNT		(NPSTRIP * 10)
#define IDLE_GLOW_CNT		(NELEM(glow_scale) * 5)

#define IDLE_HUE_STEP_RB	23
#define IDLE_HUE_STEP_BAR	9

static uint8_t idle_anim_state;	// Current idle show to display
static uint16_t idle_cnt;	// Counter to track animation duration
static uint16_t idle_hue;	// Persistent cyclic hue

/*
 * Animation statemachine while idle
 * Blank is the starting point. A random secondary animation is activated and
 * then returns to blank.
 */
static void idle_animate_reset(void)
{
	idle_anim_state = ST_IDLE_ANIM_BLANK;
	idle_cnt = 0;
}

static void idle_animate_step(void)
{
	uint8_t tmp;
	uint8_t dir;
	uint16_t h;

	switch(idle_anim_state) {
	case ST_IDLE_ANIM_BLANK:
		// Blackness
		if(!idle_cnt) {
			bangpower.clear();
			plate_color(0, 0, 0);
		}
		if(++idle_cnt >= IDLE_BLANK_CNT) {
			idle_cnt = 0;
			idle_anim_state = random(ST_IDLE_ANIM_RANDOM, ST_IDLE_ANIM_MAX);
		}
		break;

	case ST_IDLE_ANIM_RANDOM:
		// Random pixel colors
		for(uint8_t i = 0; i < NPIXELS; i++)
			bangpower.setPixelColorHsv(i, random(NEO_HSV_HUE_MIN, NEO_HSV_HUE_MAX+1), random(192, 256), random(64, 128));
		plate_color(random(), random(), random());
		if(++idle_cnt >= IDLE_RANDOM_CNT) {
			idle_cnt = 0;
			idle_anim_state = ST_IDLE_ANIM_BLANK;
		}
		break;
	case ST_IDLE_ANIM_PINGPONG:
		// Move a pixel from bottom/top to middle and back
		if(!idle_cnt)
			plate_color(0, 0, 0);
		tmp = idle_cnt % (NPSTRIP/2 + 3);
		dir = idle_cnt / (NPSTRIP/2 + 3);
		if(dir & 1)
			tmp = (NPSTRIP/2 + 3) - tmp - 1;
		bangpower.clear();
		if(tmp < NPSTRIP/2) {
			bangpower.setPixelColorHsv(tmp, 4 * NEO_HSV_HUE_SEXTANT - 4 * NEO_HSV_HUE_SEXTANT * tmp / NPSTRIP, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX);
			bangpower.setPixelColorHsv(NPSTRIP - tmp - 1, (4 * NEO_HSV_HUE_SEXTANT) * tmp / NPSTRIP, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX);
		} else if(tmp == NPSTRIP/2) {
			bangpower.setPixelColorHsv(tmp, 2 * NEO_HSV_HUE_SEXTANT, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX);
		} else if(tmp == NPSTRIP/2 + 1) {
			bangpower.setPixelColorHsv(NPSTRIP + 1, 2 * NEO_HSV_HUE_SEXTANT, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX/2);
			bangpower.setPixelColorHsv(NPSTRIP + 2, 2 * NEO_HSV_HUE_SEXTANT, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX/2);
		} else {
			bangpower.setPixelColorHsv(NPSTRIP + 0, 2 * NEO_HSV_HUE_SEXTANT, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX/3);
			bangpower.setPixelColorHsv(NPSTRIP + 3, 2 * NEO_HSV_HUE_SEXTANT, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX/3);
		}
		if(++idle_cnt >= IDLE_PINGPONG_CNT) {
			idle_cnt = 0;
			idle_anim_state = ST_IDLE_ANIM_BLANK;
		}
		break;
	case ST_IDLE_ANIM_RAINBOW:
		// Show rainbow colors on strip
		idle_hue += IDLE_HUE_STEP_RB;
		if(idle_hue >= NEO_HSV_HUE_MAX)
			idle_hue -= NEO_HSV_HUE_MAX;
		h = idle_hue;
		for(int8_t i = 0; i < NPSTRIP; i++) {
			if(h >= NEO_HSV_HUE_MAX)
				h -= NEO_HSV_HUE_MAX;
			bangpower.setPixelColorHsv(i, h, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX/3);
			if(i == NPSTRIP / 2) {
				bangpower.setPixelColorHsv(NPSTRIP+0, h, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX/5);
				bangpower.setPixelColorHsv(NPSTRIP+1, h, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX/4);
				bangpower.setPixelColorHsv(NPSTRIP+2, h, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX/4);
				bangpower.setPixelColorHsv(NPSTRIP+3, h, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX/5);
			}
			h += IDLE_HUE_STEP_RB;
		}
		plate_color(0, 0, 0);
		if(++idle_cnt >= IDLE_RAINBOW_CNT) {
			idle_cnt = 0;
			idle_anim_state = ST_IDLE_ANIM_BLANK;
		}
		break;
	case ST_IDLE_ANIM_BAR:
		// Random colors on the plate
		bangpower.clear();
		idle_hue += IDLE_HUE_STEP_BAR;
		if(idle_hue >= NEO_HSV_HUE_MAX)
			idle_hue -= NEO_HSV_HUE_MAX;
		// Use the HSV conversion from the strip to set the plate color
		// then clear the strip again.
		bangpower.setPixelColorHsv(NPSTRIP/2, idle_hue, NEO_HSV_SAT_MAX, NEO_HSV_VAL_MAX);
		plate_color(NPSTRIP/2);
		bangpower.setPixelColor(NPSTRIP/2, 0);
		if(++idle_cnt >= IDLE_BAR_CNT) {
			idle_cnt = 0;
			idle_anim_state = ST_IDLE_ANIM_BLANK;
		}
		break;
	case ST_IDLE_ANIM_GLOW:
		// White glow on the plate
		if(!idle_cnt)
			bangpower.clear();
		tmp = pgm_read_byte(&glow_scale[idle_cnt % NELEM(glow_scale)]);
		plate_color(tmp, tmp, tmp);
		if(++idle_cnt >= IDLE_GLOW_CNT) {
			idle_cnt = 0;
			idle_anim_state = ST_IDLE_ANIM_BLANK;
		}
		break;
	}
	bangpower.show();
}

/*
 * Win animation
 * Blink the strip with green using quadratic scale for approx linear
 * perception.
 */
static const uint8_t win_intensities[] PROGMEM = {
	  1,   4,  12,  28, 50, 78, 113, 154, 201, 255,
	201, 154, 113,  78, 50, 28,  12,   4,   1,   0,
};

uint8_t win_step;

static void win_animate_reset(void)
{
	win_step = 0;
}

static void win_animate_step(void)
{
	uint8_t v = pgm_read_byte(&win_intensities[win_step % NELEM(win_intensities)]);
	for(uint8_t i = 0; i < NPIXELS; i++) {
		bangpower.setPixelColorHsv(i, 2*NEO_HSV_HUE_SEXTANT, NEO_HSV_SAT_MAX, v);
	}
	bangpower.show();

	// Plate color reverses from strip
	v = pgm_read_byte(&win_intensities[(win_step + NELEM(win_intensities)/2) % NELEM(win_intensities)]);
	plate_color(0, v, 0);

	win_step++;
	if(win_step >= NELEM(win_intensities))
		win_step = 0;
}

/*
 * State-change and entry/exit handling of the statemachine
 */
static void state_set(uint8_t newstate)
{
#ifdef DEBUG
	Serial.print(F("state_set: "));
	Serial.print(state);
	Serial.print(F(" --> "));
	Serial.println(newstate);
	Serial.flush();
#endif
	// State exit handling
	switch(state) {
	case ST_IDLE:
		timer = 0;
		break;

	case ST_PEAKMEASURE:
		timer_blink = 0;
		timer = 0;
		break;

	case ST_PEAKFAIL:
		timer_blink = 0;
		plate_color(0xff, 0x00, 0x00);
		break;
	}

	state = newstate;

	// State entry handling
	switch(state) {
	case ST_IDLE:
		peak_clear();
		idle_animate_reset();
		timer = TIME_IDLEANIM;
		break;

	case ST_PEAKMEASURE:
		bangpower.clear();
		bangpower.show();
		peakpp = PEAKMIN;
		timer_blink = TIME_PLATEBLINK;
		timer = TIME_PEAKMEASURE;
		break;

	case ST_PEAKFAIL:
		peak_clear();
		sound_tone(NOTE_C3, DUR_1_4);	// Beep
		timer_blink = TIME_PLATEBLINKFAIL;
		blinkstate = 0;
		bangpower.clear();
		bangpower.show();
		break;

	case ST_PEAKWAIT:
		timer = TIME_PEAKEXTEND;
		break;

	case ST_PEAKDONE:
		peak_clear();
		sound_tone(NOTE_C5, DUR_1_8);	// Beep
#ifdef DEBUG
		Serial.print(F("Peak: "));
		Serial.print(peakl);
		Serial.write(' ');
		Serial.print(peakh);
		Serial.write(' ');
		Serial.println(peakpp);
#endif
		peak_animate_reset();
		timer = TIME_PEAKANIM;
		break;

	case ST_WINNER:
		timer = TIME_WINANIM;
		win_animate_reset();
		sound_tune_start();
		break;
	}
}

/*
 * Arduino hardware setup
 */
void setup()
{
	PORTB = 0;			// Clear all I/O pins
	PORTD = 0;
	MCUCR |= _BV(PUD);		// Globally disable pull-ups

	pinMode(PIN_HANDS, INPUT);	// Setup I/O direction
	digitalWrite(PIN_PULL, 1);	// Disable the high-side driver
	pinMode(PIN_PULL, OUTPUT);
	digitalWrite(PIN_TONE, 1);	// Enable the low-side driver
	pinMode(PIN_TONE, OUTPUT);
	pinMode(PIN_LEDR, OUTPUT);
	pinMode(PIN_LEDG, OUTPUT);
	pinMode(PIN_LEDB, OUTPUT);

	pinMode(PIN_UNUSED4, OUTPUT);	// Make unused pins output (are set to 0)
	pinMode(PIN_UNUSED7, OUTPUT);
	pinMode(PIN_UNUSED8, OUTPUT);
	pinMode(PIN_UNUSED9, OUTPUT);
	pinMode(PIN_UNUSED11, OUTPUT);

	analogWrite(PIN_LEDR, 0x7f);	// Write a color to the plate
	analogWrite(PIN_LEDG, 0x3f);	// to initialize the PWM
	analogWrite(PIN_LEDB, 0x1f);

	sound_init();
	sound_off();
	peak_init();
	adc_init();

	bangpower.begin();

	Serial.begin(115200);
	delay(10);		// Let the serial line stabelize before we send data
	Serial.print(F("Bang Head Here!\r\nHitting your head to satisfy your needs\r\n(c) 2016 Vagrearg\r\nDistributed under GPLv3\r\n\r\n"));

	prevtime = millis();
	state_set(ST_IDLE);
	handsold = hands = digitalRead(PIN_HANDS);
	timer_hands = 0;

	timer_dog = TIME_WATCHDOG;	// Setup watchdog kick timer
	wdt_disable();			// Disable the watchdog
	_NOP();
	wdt_enable(WDTO_250MS);		// Re-enable the watchdog at 250ms
}

/*
 * Main program loop
 */
#define __ev(mask)	(events & (mask))

void loop()
{
	// Read the peak values and restart next measurement
	// This is done every time we loop to ensure fastest sample-rate possible.
	// The sample rate is not constant in time, but we do not care too
	// much. The only thing important is that we get the peak detection as
	// good as possible. The hardware helps to retain the peak levels.
	//
	// The placement of this code at the start of the loop() function is to
	// ensure maximum async overlap between the measurement and the rest of
	// the code.
	if(peak_isenabled() && adc_isready()) {
		uint8_t low = ADCL;
		int16_t val = (((uint16_t)ADCH << 8) | low) - 512;
		if(ADMUX & _BV(MUX0)) {
			if(val < peakl)
				peakl = val;
		} else {
			if(val > peakh)
				peakh = val;
		}
		adc_measure_start();		// Start new measurement asap
		peakpp = peakh - peakl;		// Recalculate the p-p value
	}

	uint8_t now = millis();
	uint8_t tdiff = now - prevtime;
	uint8_t events = 0;

	if(tdiff) {
		// At least 1ms has passed
		prevtime = now;

		// Debounce the hands on the plate and generate events for it
		// at the end of the debounce period
		uint8_t handsnew = digitalRead(PIN_HANDS);
		if(handsold != handsnew) {
			// Restart the timer each time a change is detected
			handsold = handsnew;
			timer_hands = TIME_HANDS;
		} else {
			if(timer_hands) {
				if(timer_hands <= tdiff) {
					// Debounce timer expired, save state and generate event
					timer_hands = 0;
					// Only generate an event if the new state actually differs from the last state
					if(hands != handsnew) {
						hands = handsnew;
						events |= handsnew ? EV_HANDS_OFF : EV_HANDS_ON;
					}
				} else
					timer_hands -= tdiff;
			}
		}

		// Handle timer events
		events |= do_timer(tdiff, timer_tone, EV_TMRTONE);
		events |= do_timer(tdiff, timer_peak, EV_TMRPEAK);
		events |= do_timer(tdiff, timer_blink, EV_TMRBLINK);
		events |= do_timer(tdiff, timer, EV_TIMER);

		if(do_timer(tdiff, timer_dog, 1)) {
			wdt_reset();
			timer_dog = TIME_WATCHDOG;
		}
	}

#ifdef DEBUG
	if(__ev(EV_HANDS_OFF))
		Serial.println(F("Hands off"));
	if(__ev(EV_HANDS_ON))
		Serial.println(F("Hands on"));
#endif

	// Tone timing is a general property across states
	// Just as well do it here immediately
	if(__ev(EV_TMRTONE))
		sound_off();

	// Run the statemachine
	switch(state) {
	case ST_IDLE:
		/* Idle state; wait for something to happen, else animate lights */
		if(__ev(EV_HANDS_ON)) {
			state_set(ST_HANDS_ON);
		} else if(__ev(EV_TIMER)) {
			timer = TIME_IDLEANIM;
			idle_animate_step();
		}
		break;
/*
 * States og the game:
 * HANDS_ON		Detected hands on, prepare and beep
 * WAIT_TONE		Hands on, wait for beep to finish
 * WAIT_PEAKENABLE	Hands on, wait for enabled peak measurement
 * PEAK_MEASURE		Hands on, now you may bang your head
 */
	case ST_HANDS_ON:
		/* Hands are placed on plate, start process to bang */

		// If the hand-plates are released, goto idle immediately
		if(__ev(EV_HANDS_OFF)) {
			state_set(ST_IDLE);
			break;
		}

		sound_tone(NOTE_G3, DUR_1_8);	// Beep
		bangpower.clear();
		for(uint8_t i = 0; i < NPSTRIP/2; i++) {
			bangpower.setPixelColor(i, 0, 0xff >> (NPSTRIP/2 - i), 0xff >> i);
			bangpower.setPixelColor(NPSTRIP - i - 1, 0xff >> i, 0xff >> (NPSTRIP/2 - i), 0);
		}
		bangpower.setPixelColor(NPSTRIP+0, 0, 0x3f, 0);
		bangpower.setPixelColor(NPSTRIP+1, 0, 0x7f, 0);
		bangpower.setPixelColor(NPSTRIP/2, 0, 0xff, 0);
		bangpower.setPixelColor(NPSTRIP+2, 0, 0x7f, 0);
		bangpower.setPixelColor(NPSTRIP+3, 0, 0x3f, 0);
		bangpower.show();
		plate_color(0x7f, 0x7f, 0x7f);
		blinkstate = 1;
		state_set(ST_WAIT_TONE);
		break;

	case ST_WAIT_TONE:
		/* Wait for the beep to end */

		// If the hand-plates are released, goto idle immediately
		if(__ev(EV_HANDS_OFF) || hands) {
			sound_tone(NOTE_F3, DUR_1_32);
			state_set(ST_IDLE);
		} else if(__ev(EV_TMRTONE)) {
			peak_enable();	// This starts the timer_peak
			state_set(ST_WAIT_PEAKENABLE);
		}
		break;

	case ST_WAIT_PEAKENABLE:
		if(__ev(EV_HANDS_OFF) || hands) {
			sound_tone(NOTE_F3, DUR_1_32);
			state_set(ST_IDLE);
		} else if(__ev(EV_TMRPEAK)) {
			// Peak detector ready to be enabled for read
			peak_enable_input();
			state_set(ST_PEAKMEASURE);
		}
		break;

	case ST_PEAKMEASURE:
		/* Waiting for a head bang/peak detection */
		if(__ev(EV_HANDS_OFF) || hands) {
			sound_tone(NOTE_F3, DUR_1_32);
			state_set(ST_IDLE);
			break;
		}

		if(__ev(EV_TMRBLINK)) {
			// Blink the plate while measuring
			timer_blink = TIME_PLATEBLINK;	// Set for next blink
			if(blinkstate & 1)
				plate_color(0x00, 0x00, 0x00);
			else
				plate_color(0xff, 0xff, 0xff);
			blinkstate++;
		}
		if(__ev(EV_TIMER)) {
			// Timeout, no head banged the plate
			plate_color(0xff, 0x00, 0x00);
			state_set(ST_PEAKFAIL);
		} else if(peakpp >= PEAKMIN) {
			// We received a Bang!
			plate_color(0x00, 0x00, 0xff);
			state_set(ST_PEAKWAIT);
		}
		break;

	case ST_PEAKFAIL:
		if(__ev(EV_TMRBLINK)) {
			timer_blink = TIME_PLATEBLINKFAIL;	// Set for next blink
			if(blinkstate & 1)
				plate_color(0x00, 0x00, 0x00);
			else
				plate_color(0xff, 0x00, 0x00);
			if(++blinkstate > NFAILBLINK)
				state_set(!hands ? ST_WAIT_HANDS_OFF : ST_IDLE);
		}
		break;

	case ST_PEAKWAIT:
		/* A bang was detected, wait for the bang to fade out */
		if(__ev(EV_TIMER)) {
			state_set(ST_PEAKDONE);
		}
		break;

	case ST_PEAKDONE:
		// peakpp now holds the maximum extent of the Bang!
		if(__ev(EV_TIMER)) {
			timer = TIME_PEAKANIM;
			if(!peak_animate_step() && hands)
				state_set(ST_WAIT_HANDS_OFF);
		}
		break;

	case ST_WAIT_HANDS_OFF:
		// Wait for hands to be released
		if(__ev(EV_HANDS_OFF) || hands) {
			state_set(ST_IDLE);
		}
		break;

	case ST_WINNER:
		// Bang is dead center, play tune and blink green
		if(__ev(EV_TMRTONE)) {
			if(!sound_tune_next())
				state_set(ST_WAIT_HANDS_OFF);
		}
		if(__ev(EV_TIMER)) {
			timer = TIME_WINANIM;
			win_animate_step();
		}
		break;

	default:
		// Catch all if something went wrong
		// Not sure if this is appropriate. A soft reset may be better
		// because the controller's state may be corrupted in other
		// ways if we arrive here. We may just as well disable
		// interrupts and loop in a while until the watchdog resets us.
		state_set(ST_IDLE);
		break;
	}
}
